{-# LANGUAGE DataKinds, TypeFamilies, TypeOperators, FlexibleInstances, OverloadedStrings, ViewPatterns #-}
{-# LANGUAGE RecordWildCards, GeneralizedNewtypeDeriving, DeriveTraversable, FlexibleContexts, DeriveGeneric #-}
{-# OPTIONS_GHC -fno-warn-unused-binds -fno-warn-unused-imports -fcontext-stack=370 #-}
module 1Gb.API (
  -- * Client and Server
  ServerConfig(..),
  1GbBackend,
  create1GbClient,
  run1GbServer,
  run1GbClient,
  run1GbClientWithManager,
  1GbClient,
  -- ** Servant
  1GbAPI,
  ) where

import 1Gb.Types

import Data.Aeson (Value)
import Data.Coerce (coerce)
import Servant.API
import Servant (serve, ServantErr)
import Web.HttpApiData
import qualified Network.Wai.Handler.Warp as Warp
import qualified Data.Text as T
import Data.Text (Text)
import Servant.Common.BaseUrl(BaseUrl(..))
import Servant.Client (ServantError, client, Scheme(Http))
import Data.Proxy (Proxy(..))
import Control.Monad.IO.Class
import Data.Function ((&))
import GHC.Exts (IsString(..))
import qualified Data.Map as Map
import GHC.Generics (Generic)
import Data.Monoid ((<>))
import Servant.API.Verbs (Verb, StdMethod(..))
import Control.Monad.Except (ExceptT)
import Network.HTTP.Client (Manager, newManager, defaultManagerSettings)
import Network.HTTP.Types.Method (methodOptions)

instance ReflectMethod 'OPTIONS where
  reflectMethod _ = methodOptions




-- For the form data code generation.
lookupEither :: FromHttpApiData b => Text -> [(Text, Text)] -> Either Text b
lookupEither key assocs =
  case lookup key assocs of
    Nothing -> Left $ "Could not find parameter " <> key <> " in form data"
    Just value -> parseQueryParam value

-- | Servant type-level API, generated from the Swagger spec for 1Gb.
type 1GbAPI
    =    "balance" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] Double -- 'balanceGet' route
    :<|> "balance.lic" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] Double -- 'balanceLicGet' route
    :<|> "billing" :> "debt" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200] -- 'billingDebtGet' route
    :<|> "billing" :> "info" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200_1] -- 'billingInfoGet' route
    :<|> "billing" :> "limits" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] ( ) -- 'billingLimitsGet' route
    :<|> "dns" :> "prolong" :> "list" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200_4] -- 'dnsProlongListGet' route
    :<|> "email" :> "add" :> QueryParam "_token_" Text :> QueryParam "add_email" Text :> Verb 'GET 200 '[JSON] Text -- 'emailAddGet' route
    :<|> "email" :> "alias" :> "delete" :> QueryParam "_token_" Text :> QueryParam "_key_" Text :> Verb 'GET 200 '[JSON] Text -- 'emailAliasDeleteGet' route
    :<|> "email" :> "alias" :> "list" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200_5] -- 'emailAliasListGet' route
    :<|> "email" :> "delete" :> QueryParam "_token_" Text :> QueryParam "_key_" Text :> Verb 'GET 200 '[JSON] Text -- 'emailDeleteGet' route
    :<|> "email" :> "list" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200_6] -- 'emailListGet' route
    :<|> "updates" :> "last" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] Int -- 'updatesLastGet' route
    :<|> "updates" :> "list" :> QueryParam "_token_" Text :> QueryParam "_key_" Int :> Verb 'GET 200 '[JSON] [Inline_response_200_10] -- 'updatesListGet' route
    :<|> "website" :> "delete" :> QueryParam "_token_" Text :> QueryParam "_key_" Text :> Verb 'GET 200 '[JSON] Text -- 'websiteDeleteGet' route
    :<|> "website" :> "list" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200_14] -- 'websiteListGet' route
    :<|> "auth" :> "login" :> QueryParam "login" Text :> QueryParam "salt" Text :> QueryParam "response" Text :> Verb 'POST 200 '[JSON] Text -- 'authLoginPost' route
    :<|> "auth" :> "start" :> QueryParam "login" Text :> Verb 'GET 200 '[JSON] Text -- 'authStartGet' route
    :<|> "dns" :> "add" :> QueryParam "_token_" Text :> QueryParam "add_dns_zone" Text :> QueryParam "add_dns_record" Text :> QueryParam "iis_site_for_dns" Text :> QueryParam "bind_only" Bool :> Verb 'GET 200 '[JSON] Text -- 'dnsAddGet' route
    :<|> "dns" :> "bind" :> QueryParam "_token_" Text :> QueryParam "_key_" Int :> QueryParam "bind_to" Text :> Verb 'GET 200 '[JSON] Text -- 'dnsBindGet' route
    :<|> "dns" :> "list" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200_3] -- 'dnsListGet' route
    :<|> "dns" :> "prolong" :> "list" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200_4] -- 'dnsProlongListGet' route
    :<|> "dns" :> "raw" :> QueryParam "_token_" Text :> QueryParam "_key_" Int :> QueryParam "s_del" Int :> QueryParam "s_add" Int :> QueryParam "dns_type" Text :> QueryParam "dns_name" Text :> QueryParam "dns_value" Text :> Verb 'GET 200 '[JSON] Text -- 'dnsRawGet' route
    :<|> "dns" :> "raw" :> "managedmx" :> QueryParam "_token_" Text :> QueryParam "_key_" Int :> QueryParam "managed_mx" Int :> QueryParam "upd_managed_mx" Int :> Verb 'GET 200 '[JSON] Text -- 'dnsRawManagedmxGet' route
    :<|> "ftp" :> "list" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200_7] -- 'ftpListGet' route
    :<|> "iis" :> "list" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200_8] -- 'iisListGet' route
    :<|> "mssql" :> "list" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200_9] -- 'mssqlListGet' route
    :<|> "database" :> "list" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200_2] -- 'databaseListGet' route
    :<|> "vds" :> "create" :> QueryParam "_token_" Text :> QueryParam "type" Text :> QueryParam "cr_hvmem" Int :> QueryParam "cr_hvdsk" Int :> QueryParam "cr_hvcpu" Int :> QueryParam "template" Text :> QueryParam "force_memo" Text :> QueryParam "hv_1gb_panel_version" Text :> QueryParam "panel_install" Bool :> QueryParam "cr_ssd1" Bool :> QueryParam "cr_hv_all_system" Bool :> Verb 'GET 200 '[JSON] Text -- 'vdsCreateGet' route
    :<|> "vds" :> "delete" :> QueryParam "_token_" Text :> QueryParam "_key_" Int :> QueryParam "force_critical" Bool :> Verb 'GET 200 '[JSON] Text -- 'vdsDeleteGet' route
    :<|> "vds" :> "ip" :> "add" :> QueryParam "_token_" Text :> QueryParam "_key_" Int :> Verb 'GET 200 '[JSON] Text -- 'vdsIpAddGet' route
    :<|> "vds" :> "ip" :> "delete" :> QueryParam "_token_" Text :> QueryParam "_key_" Int :> QueryParam "del_ip" Text :> Verb 'GET 200 '[JSON] Text -- 'vdsIpDeleteGet' route
    :<|> "vds" :> "ip" :> "list" :> QueryParam "_token_" Text :> QueryParam "_key_" Int :> Verb 'GET 200 '[JSON] [Inline_response_200_11] -- 'vdsIpListGet' route
    :<|> "vds" :> "list" :> QueryParam "_token_" Text :> QueryParam "_key_" Int :> Verb 'GET 200 '[JSON] [Inline_response_200_12] -- 'vdsListGet' route
    :<|> "vds" :> "reboot" :> QueryParam "_token_" Text :> QueryParam "_key_" Int :> Verb 'GET 200 '[JSON] Text -- 'vdsRebootGet' route
    :<|> "vds" :> "templates-list" :> QueryParam "_token_" Text :> Verb 'GET 200 '[JSON] [Inline_response_200_13] -- 'vdsTemplatesListGet' route
    :<|> "vds" :> "update" :> QueryParam "_token_" Text :> QueryParam "_key_" Int :> QueryParam "cr_hvmem" Int :> QueryParam "cr_hvdsk" Int :> QueryParam "cr_hvcpu" Int :> Verb 'GET 200 '[JSON] Text -- 'vdsUpdateGet' route

-- | Server or client configuration, specifying the host and port to query or serve on.
data ServerConfig = ServerConfig {
    configHost :: String,  -- ^ Hostname to serve on, e.g. "127.0.0.1"
    configPort :: Int      -- ^ Port to serve on, e.g. 8080
  } deriving (Eq, Ord, Show, Read)

-- | List of elements parsed from a query.
newtype QueryList (p :: CollectionFormat) a = QueryList { fromQueryList :: [a] }
  deriving (Functor, Applicative, Monad, Foldable, Traversable)

-- | Formats in which a list can be encoded into a HTTP path.
data CollectionFormat = CommaSeparated -- ^ CSV format for multiple parameters.
                      | SpaceSeparated -- ^ Also called "SSV"
                      | TabSeparated -- ^ Also called "TSV"
                      | PipeSeparated -- ^ `value1|value2|value2`
                      | MultiParamArray -- ^ Using multiple GET parameters, e.g. `foo=bar&foo=baz`. Only for GET params.

instance FromHttpApiData a => FromHttpApiData (QueryList 'CommaSeparated a) where
    parseQueryParam = parseSeparatedQueryList ','

instance FromHttpApiData a => FromHttpApiData (QueryList 'TabSeparated a) where
    parseQueryParam = parseSeparatedQueryList '\t'

instance FromHttpApiData a => FromHttpApiData (QueryList 'SpaceSeparated a) where
    parseQueryParam = parseSeparatedQueryList ' '

instance FromHttpApiData a => FromHttpApiData (QueryList 'PipeSeparated a) where
    parseQueryParam = parseSeparatedQueryList '|'

instance FromHttpApiData a => FromHttpApiData (QueryList 'MultiParamArray a) where
    parseQueryParam = error "unimplemented FromHttpApiData for MultiParamArray collection format"

parseSeparatedQueryList :: FromHttpApiData a => Char -> Text -> Either Text (QueryList p a)
parseSeparatedQueryList char = fmap QueryList . mapM parseQueryParam . T.split (== char)

instance ToHttpApiData a => ToHttpApiData (QueryList 'CommaSeparated a) where
    toQueryParam = formatSeparatedQueryList ','

instance ToHttpApiData a => ToHttpApiData (QueryList 'TabSeparated a) where
    toQueryParam = formatSeparatedQueryList '\t'

instance ToHttpApiData a => ToHttpApiData (QueryList 'SpaceSeparated a) where
    toQueryParam = formatSeparatedQueryList ' '

instance ToHttpApiData a => ToHttpApiData (QueryList 'PipeSeparated a) where
    toQueryParam = formatSeparatedQueryList '|'

instance ToHttpApiData a => ToHttpApiData (QueryList 'MultiParamArray a) where
    toQueryParam = error "unimplemented ToHttpApiData for MultiParamArray collection format"

formatSeparatedQueryList :: ToHttpApiData a => Char ->  QueryList p a -> Text
formatSeparatedQueryList char = T.intercalate (T.singleton char) . map toQueryParam . fromQueryList


-- | Backend for 1Gb.
-- The backend can be used both for the client and the server. The client generated from the 1Gb Swagger spec
-- is a backend that executes actions by sending HTTP requests (see @create1GbClient@). Alternatively, provided
-- a backend, the API can be served using @run1GbServer@.
data 1GbBackend m = 1GbBackend {
    balanceGet :: Maybe Text -> m Double{- ^ Возвращает текущий баланс аккаунта -},
    balanceLicGet :: Maybe Text -> m Double{- ^ Возвращает текущий баланс аккаунта лицензий -},
    billingDebtGet :: Maybe Text -> m [Inline_response_200]{- ^ Данный запрос позволяет получить информацию по планируемым датам наступления негативного баланса и отключения -},
    billingInfoGet :: Maybe Text -> m [Inline_response_200_1]{- ^ Этот запрос позволяет получить информацию об аккаунте, а также базовую статистику использования ресурсов. -},
    billingLimitsGet :: Maybe Text -> m ( ){- ^  -},
    dnsProlongListGet :: Maybe Text -> m [Inline_response_200_4]{- ^  -},
    emailAddGet :: Maybe Text -> Maybe Text -> m Text{- ^  -},
    emailAliasDeleteGet :: Maybe Text -> Maybe Text -> m Text{- ^  -},
    emailAliasListGet :: Maybe Text -> m [Inline_response_200_5]{- ^  -},
    emailDeleteGet :: Maybe Text -> Maybe Text -> m Text{- ^  -},
    emailListGet :: Maybe Text -> m [Inline_response_200_6]{- ^  -},
    updatesLastGet :: Maybe Text -> m Int{- ^  -},
    updatesListGet :: Maybe Text -> Maybe Int -> m [Inline_response_200_10]{- ^  -},
    websiteDeleteGet :: Maybe Text -> Maybe Text -> m Text{- ^  -},
    websiteListGet :: Maybe Text -> m [Inline_response_200_14]{- ^  -},
    authLoginPost :: Maybe Text -> Maybe Text -> Maybe Text -> m Text{- ^ Возвращает ключ доступа, с которым в дальнейшем будут осуществляться всё взаимодействие с API  (параметр 'token'). Для вычисления значения параметра 'response' необходимо получить представленную в hex MD5-сумму от  объединения строк пароля пользователя, полученной соли и символа переноса строки. Псевдокод операции:  hex(md5(password&nbsp;+&nbsp;salt&nbsp;+&nbsp;'\\n')) -},
    authStartGet :: Maybe Text -> m Text{- ^ Возвращает соль для авторизации. Это первый этап идентификации пользователя. Затем соль используется  во втором, окончательном этапе авторизации, в запросе /auth/login. Запрос не требует дополнительных параметров. -},
    dnsAddGet :: Maybe Text -> Maybe Text -> Maybe Text -> Maybe Text -> Maybe Bool -> m Text{- ^  -},
    dnsBindGet :: Maybe Text -> Maybe Int -> Maybe Text -> m Text{- ^  -},
    dnsListGet :: Maybe Text -> m [Inline_response_200_3]{- ^  -},
    dnsProlongListGet :: Maybe Text -> m [Inline_response_200_4]{- ^  -},
    dnsRawGet :: Maybe Text -> Maybe Int -> Maybe Int -> Maybe Int -> Maybe Text -> Maybe Text -> Maybe Text -> m Text{- ^  -},
    dnsRawManagedmxGet :: Maybe Text -> Maybe Int -> Maybe Int -> Maybe Int -> m Text{- ^  -},
    ftpListGet :: Maybe Text -> m [Inline_response_200_7]{- ^  -},
    iisListGet :: Maybe Text -> m [Inline_response_200_8]{- ^  -},
    mssqlListGet :: Maybe Text -> m [Inline_response_200_9]{- ^  -},
    databaseListGet :: Maybe Text -> m [Inline_response_200_2]{- ^ Возвращает список зарегистрированных аккаунтов в базах данных MySQL и Postgres. Для вычисления имени домена необходимо объединить имя базы данных,   идентификатор сервера и суффикс (поле url_suffix). Например, если ID сервера равен 123 (значит это MySQL — см. описание поля 'server' в ответе), а суффикс равен '1gb.ru', то домен   для доступа будет равен 'mysql123.1gb.ru'. -},
    vdsCreateGet :: Maybe Text -> Maybe Text -> Maybe Int -> Maybe Int -> Maybe Int -> Maybe Text -> Maybe Text -> Maybe Text -> Maybe Bool -> Maybe Bool -> Maybe Bool -> m Text{- ^ Изменение параметров уже существующего VDS -},
    vdsDeleteGet :: Maybe Text -> Maybe Int -> Maybe Bool -> m Text{- ^ Предоставляет функционал для удаления VDS -},
    vdsIpAddGet :: Maybe Text -> Maybe Int -> m Text{- ^ Этот вызов позволяет добавить ещё один IP адрес на указанный VDS -},
    vdsIpDeleteGet :: Maybe Text -> Maybe Int -> Maybe Text -> m Text{- ^ Этот вызов позволяет отвязать IP адрес от VDS -},
    vdsIpListGet :: Maybe Text -> Maybe Int -> m [Inline_response_200_11]{- ^ Список IP-адресов виртуальных серверов -},
    vdsListGet :: Maybe Text -> Maybe Int -> m [Inline_response_200_12]{- ^ Позволяет получить список виртуальных серверов -},
    vdsRebootGet :: Maybe Text -> Maybe Int -> m Text{- ^ Предоставляет функционал для перезагрузки VDS -},
    vdsTemplatesListGet :: Maybe Text -> m [Inline_response_200_13]{- ^ Позволяет получить список шаблонов виртуальных серверов -},
    vdsUpdateGet :: Maybe Text -> Maybe Int -> Maybe Int -> Maybe Int -> Maybe Int -> m Text{- ^ Создание новых VDS с произвольными характеристиками -}
  }

newtype 1GbClient a = 1GbClient { runClient :: Manager -> BaseUrl -> ExceptT ServantError IO a }
    deriving Functor

instance Applicative 1GbClient where
    pure x = 1GbClient (\_ _ -> pure x)
    (1GbClient f) <*> (1GbClient x) = 1GbClient (\manager url -> f manager url <*> x manager url)

instance Monad 1GbClient where
    (1GbClient a) >>= f = 1GbClient (\manager url -> do
        value <- a manager url
        runClient (f value) manager url)

instance MonadIO 1GbClient where
    liftIO io = 1GbClient (\_ _ -> liftIO io)

create1GbClient :: 1GbBackend 1GbClient
create1GbClient = 1GbBackend{..}
  where
    ((coerce -> balanceGet) :<|>
     (coerce -> balanceLicGet) :<|>
     (coerce -> billingDebtGet) :<|>
     (coerce -> billingInfoGet) :<|>
     (coerce -> billingLimitsGet) :<|>
     (coerce -> dnsProlongListGet) :<|>
     (coerce -> emailAddGet) :<|>
     (coerce -> emailAliasDeleteGet) :<|>
     (coerce -> emailAliasListGet) :<|>
     (coerce -> emailDeleteGet) :<|>
     (coerce -> emailListGet) :<|>
     (coerce -> updatesLastGet) :<|>
     (coerce -> updatesListGet) :<|>
     (coerce -> websiteDeleteGet) :<|>
     (coerce -> websiteListGet) :<|>
     (coerce -> authLoginPost) :<|>
     (coerce -> authStartGet) :<|>
     (coerce -> dnsAddGet) :<|>
     (coerce -> dnsBindGet) :<|>
     (coerce -> dnsListGet) :<|>
     (coerce -> dnsProlongListGet) :<|>
     (coerce -> dnsRawGet) :<|>
     (coerce -> dnsRawManagedmxGet) :<|>
     (coerce -> ftpListGet) :<|>
     (coerce -> iisListGet) :<|>
     (coerce -> mssqlListGet) :<|>
     (coerce -> databaseListGet) :<|>
     (coerce -> vdsCreateGet) :<|>
     (coerce -> vdsDeleteGet) :<|>
     (coerce -> vdsIpAddGet) :<|>
     (coerce -> vdsIpDeleteGet) :<|>
     (coerce -> vdsIpListGet) :<|>
     (coerce -> vdsListGet) :<|>
     (coerce -> vdsRebootGet) :<|>
     (coerce -> vdsTemplatesListGet) :<|>
     (coerce -> vdsUpdateGet)) = client (Proxy :: Proxy 1GbAPI)

-- | Run requests in the 1GbClient monad.
run1GbClient :: ServerConfig -> 1GbClient a -> ExceptT ServantError IO a
run1GbClient clientConfig cl = do
  manager <- liftIO $ newManager defaultManagerSettings
  run1GbClientWithManager manager clientConfig cl

-- | Run requests in the 1GbClient monad using a custom manager.
run1GbClientWithManager :: Manager -> ServerConfig -> 1GbClient a -> ExceptT ServantError IO a
run1GbClientWithManager manager clientConfig cl =
  runClient cl manager $ BaseUrl Http (configHost clientConfig) (configPort clientConfig) ""

-- | Run the 1Gb server at the provided host and port.
run1GbServer :: MonadIO m => ServerConfig -> 1GbBackend (ExceptT ServantErr IO)  -> m ()
run1GbServer ServerConfig{..} backend =
  liftIO $ Warp.runSettings warpSettings $ serve (Proxy :: Proxy 1GbAPI) (serverFromBackend backend)

  where
    warpSettings = Warp.defaultSettings & Warp.setPort configPort & Warp.setHost (fromString configHost)
    serverFromBackend 1GbBackend{..} =
      (coerce balanceGet :<|>
       coerce balanceLicGet :<|>
       coerce billingDebtGet :<|>
       coerce billingInfoGet :<|>
       coerce billingLimitsGet :<|>
       coerce dnsProlongListGet :<|>
       coerce emailAddGet :<|>
       coerce emailAliasDeleteGet :<|>
       coerce emailAliasListGet :<|>
       coerce emailDeleteGet :<|>
       coerce emailListGet :<|>
       coerce updatesLastGet :<|>
       coerce updatesListGet :<|>
       coerce websiteDeleteGet :<|>
       coerce websiteListGet :<|>
       coerce authLoginPost :<|>
       coerce authStartGet :<|>
       coerce dnsAddGet :<|>
       coerce dnsBindGet :<|>
       coerce dnsListGet :<|>
       coerce dnsProlongListGet :<|>
       coerce dnsRawGet :<|>
       coerce dnsRawManagedmxGet :<|>
       coerce ftpListGet :<|>
       coerce iisListGet :<|>
       coerce mssqlListGet :<|>
       coerce databaseListGet :<|>
       coerce vdsCreateGet :<|>
       coerce vdsDeleteGet :<|>
       coerce vdsIpAddGet :<|>
       coerce vdsIpDeleteGet :<|>
       coerce vdsIpListGet :<|>
       coerce vdsListGet :<|>
       coerce vdsRebootGet :<|>
       coerce vdsTemplatesListGet :<|>
       coerce vdsUpdateGet)
